import 'dart:async';
import 'event.dart';

/// A mutation to a new [State] from the given [event].
typedef FutureOr<State> Mutation(Event event);

/// Represents a state that await for an [event] to
/// mutate to a new [State].
class State {
  State();

  /// A state that cannot mutate because it has 
  /// no mutations.
  bool get isLast => this._mutations.isEmpty;

  Map<Type, Mutation> _mutations = {};

  /// Indicates whether this state can mutate from the given [event].
  bool canMutate(Event event) => this._mutations.containsKey(event?.runtimeType);

  /// Indicates whether this state can mutate from the given
  /// [T] event type.
  bool hasMutation<T extends Event>() => this._mutations.containsKey(T);

  /// Register a [mutation] when a [T] event is raised.
  void register<T extends Event>(FutureOr<State> mutation(T e)) {
    this._mutations[T] = (e) => mutation(e as T);
  }

  /// Mutate to a new [State] from the given [event].
  /// 
  /// If [event] is null or this state cannot mutate from the 
  /// given [event], an [Exception] is raised.
  Future<State> mutate(Event event) async {
    if (event == null) {
      throw Exception("empty event");
    }

    final t = event.runtimeType;
    
    if (!_mutations.containsKey(t)) {
      throw Exception("state is already mutating");
    }

    final mutation = _mutations[t];
    final newState = await mutation(event);
    return newState;
  }

  @override
  String toString() {
    final result = StringBuffer();
    result.write("[$runtimeType] < ");
    if(this.isLast) {
      result.write("end");
    }
    else {
      result.write("(");
      result.write(
          this._mutations.keys.map((t) => t.toString()).join("|"));
      result.write(")");
    }
    return result.toString();
  }
}
